/** @file sdl_screenmode.cpp

	Skelton for retropc emulator

	@author Sasaji
	@date   2019.10.01

	@brief [ screen mode ]
*/

#include "../screenmode.h"
#include <SDL.h>
#include "../../emu.h"


ScreenMode::ScreenMode()
{
	disp_device_count = 0;
	monitor_count = 0;
	screen_mode_count = 0;
}

ScreenMode::~ScreenMode()
{
}

void ScreenMode::Enum()
{
	int w, h, bpp;
	GetDesktopSize(&w, &h, &bpp);
	Enum(w, h, bpp);
}

void ScreenMode::Enum(int desktop_width, int desktop_height, int bits_per_pixel)
{
	// enumerate display device
	int max_desktop_width = 0;
	int max_desktop_height = 0;

	// enumerate screen mode for fullscreen
	int w,h,bpp;
	screen_mode_count = 0;

#ifndef USE_SDL2
	SDL_Rect **list = NULL;

	uint32_t flags = SDL_FULLSCREEN;

	disp_device_count = 1;
	disp_device[0].re.x = 0;
	disp_device[0].re.y = 0;
	disp_device[0].re.w = desktop_width;
	disp_device[0].re.h = desktop_height;

	const SDL_VideoInfo *video = SDL_GetVideoInfo();
#else
	SDL_DisplayMode mode;
	disp_device_count = SDL_GetNumVideoDisplays();
	if (disp_device_count > DISP_DEVICE_MAX) {
		disp_device_count = DISP_DEVICE_MAX;
	}
	for(int i = 0; i < disp_device_count && i < DISP_DEVICE_MAX; i++) {
		disp_device[i].name.SetN(SDL_GetDisplayName(i));
		SDL_Rect re;
		SDL_GetDisplayBounds(i, &re);
		RECT_IN(disp_device[i].re, re.x, re.y, re.w, re.h);
	}
#endif

	for(int disp_no = 0; disp_no < disp_device_count && screen_mode_count < SCREEN_MODE_MAX; disp_no++) {
#ifndef USE_SDL2
		list = SDL_ListModes(video->vfmt, flags);
		if (list == NULL || list == (SDL_Rect **)-1) break;
		for(int i = 0; list[i] != NULL && screen_mode_count < SCREEN_MODE_MAX; i++) {
			w = list[i]->w;
			h = list[i]->h;
			bpp = video->vfmt->BitsPerPixel;
#else
		int mode_count = SDL_GetNumDisplayModes(disp_no);
		for(int i = 0; i < mode_count && screen_mode_count < SCREEN_MODE_MAX; i++) {
			if (SDL_GetDisplayMode(disp_no, i, &mode) < 0) {
				continue;
			}
			w = mode.w;
			h = mode.h;
			bpp = SDL_BITSPERPIXEL(mode.format);
#endif
			if(w < MIN_WINDOW_WIDTH || h < MIN_WINDOW_HEIGHT || w < 640 || h < 480) {
				logging->out_debugf(_T("screen_mode:-- [%d] %dx%d %dbpp ignored"), disp_no, w, h, bpp);
				continue;
			}
			if(((w * 30 / h) != 40)
			&& ((w * 40 / h) != 50)
			&& ((w * 90 / h) != 160)
			&& ((w * 50 / h) != 80)) {
				logging->out_debugf(_T("screen_mode:-- [%d] %dx%d %dbpp ignored"), disp_no, w, h, bpp);
				continue;
			}
			if (w <= 1280 && (w != 640 && w != 800 && w != 1024 && w != 1280)) {
				logging->out_debugf(_T("screen_mode:-- [%d] %dx%d %dbpp ignored"), disp_no, w, h, bpp);
				continue;
			}
			bool found = false;
			for(int j = 0; j < screen_mode_count; j++) {
				if(screen_mode[j].disp_no == disp_no && screen_mode[j].width == w && screen_mode[j].height == h) {
					found = true;
					logging->out_debugf(_T("screen_mode:-- [%d] %dx%d %dbpp already exist"), disp_no, w, h, bpp);
					break;
				}
			}
			if(!found) {
				screen_mode[screen_mode_count].disp_no = disp_no;
				screen_mode[screen_mode_count].width = w;
				screen_mode[screen_mode_count].height = h;
				logging->out_debugf(_T("screen_mode:%2d [%d] %dx%d %dbpp"), screen_mode_count, disp_no, w, h, bpp);
				screen_mode_count++;

				if(max_desktop_width <= w) {
					max_desktop_width = w;
				}
				if(max_desktop_height <= h) {
					max_desktop_height = h;
				}
			}
		}
	}
	qsort(&screen_mode, screen_mode_count, sizeof(struct screen_mode_st), qsort_by_n_w_h);

	// if cannot get modes, add default screen size mode.
	if (screen_mode_count == 0) {
		screen_mode[screen_mode_count].disp_no = 0;
		screen_mode[screen_mode_count].width = desktop_width;
		screen_mode[screen_mode_count].height = desktop_height;
		screen_mode_count = 1;

		max_desktop_width = desktop_width;
		max_desktop_height = desktop_height;
	}
}

int ScreenMode::qsort_by_n_w_h(const void *a, const void *b)
{
	struct screen_mode_st *a1 = (struct screen_mode_st *)a;
	struct screen_mode_st *b1 = (struct screen_mode_st *)b;

	int rc = 0;
	int aa = a1->width * a1->height;
	int bb = b1->width * b1->height;
	if (!rc) rc = (a1->disp_no - b1->disp_no);
	if (!rc) rc = (bb - aa);
	return rc;
}

const screen_mode_t *ScreenMode::GetMode(int num) const
{
	if (0 <= num && num < screen_mode_count) {
		return &screen_mode[num];
	}
	return NULL;
}

int ScreenMode::FindMode(int disp_no, int width, int height) const
{
	int match = -1;
	for(int i = 0; i < screen_mode_count; i++) {
		if (screen_mode[i].disp_no == disp_no
		&&  screen_mode[i].width == width
		&&  screen_mode[i].height == height) {
			match = i;
			break;
		}
	}
	return match;
}

const disp_device_t *ScreenMode::GetDisp(int num) const
{
	if (0 <= num && num < disp_device_count) {
		return &disp_device[num];
	}
	return NULL;
}

void ScreenMode::GetDesktopSize(int *width, int *height, int *bpp)
{
	// desktop size on the primary monitor
#ifdef USE_SDL2
	/* SDL 2 */
	SDL_DisplayMode display_mode;

	memset(&display_mode, 0, sizeof(display_mode));
	if (SDL_GetDesktopDisplayMode(0, &display_mode)) {
		logging->out_logf(LOG_ERROR, _T("SDL_GetDesktopDisplayMode: %s."), SDL_GetError());
	}
	if (width) *width = display_mode.w;
	if (height) *height = display_mode.h;
	if (bpp) *bpp = SDL_BITSPERPIXEL(display_mode.format);
#else
	/* SDL 1 */
	const SDL_VideoInfo *video = NULL;

	// get current video mode
	if((video = SDL_GetVideoInfo()) == NULL) {
		logging->out_logf(LOG_ERROR, _T("SDL_GetVideoInfo: %s."), SDL_GetError());
		if (width) *width = 0;
		if (height) *height = 0;
		if (bpp) *bpp = 1;
		return;
	}
	if (width) *width = video->current_w;
	if (height) *height = video->current_h;
	if (bpp) *bpp = video->vfmt->BitsPerPixel;

#ifdef _DEBUG_LOG
	logging->out_debugf(_T("video->hw_available:%d"),video->hw_available);
	logging->out_debugf(_T("video->wm_available:%d"),video->wm_available);
	logging->out_debugf(_T("video->blit_hw:%d"),video->blit_hw);
	logging->out_debugf(_T("video->blit_hw_CC:%d"),video->blit_hw_CC);
	logging->out_debugf(_T("video->blit_hw_A:%d"),video->blit_hw_A);
	logging->out_debugf(_T("video->blit_sw:%d"),video->blit_sw);
	logging->out_debugf(_T("video->blit_sw_CC:%d"),video->blit_sw_CC);
	logging->out_debugf(_T("video->blit_sw_A:%d"),video->blit_sw_A);
	logging->out_debugf(_T("video->blit_fill:%d"),video->blit_fill);
	logging->out_debugf(_T("video->video_mem:%d"),video->video_mem);
	logging->out_debugf(_T("video->vfmt->BitsPerPixel:%d"),video->vfmt->BitsPerPixel);
	logging->out_debugf(_T("video->vfmt->BytesPerPixel:%d"),video->vfmt->BytesPerPixel);
	logging->out_debugf(_T("video->vfmt->Rmask:%x, Gmask:%x, Bmask:%x, Amask:%x"),video->vfmt->Rmask,video->vfmt->Gmask,video->vfmt->Bmask,video->vfmt->Amask);
	logging->out_debugf(_T("video->vfmt->Rshift:%d, Gshift:%d, Bshift:%d, Ashift:%d"),video->vfmt->Rshift,video->vfmt->Gshift,video->vfmt->Bshift,video->vfmt->Ashift);
	logging->out_debugf(_T("video->vfmt->Rloss:%d, Gloss:%d, Bloss:%d, Aloss:%d"),video->vfmt->Rloss,video->vfmt->Gloss,video->vfmt->Bloss,video->vfmt->Aloss);
	logging->out_debugf(_T("video->vfmt->colorkey:%d"),video->vfmt->colorkey);
	logging->out_debugf(_T("video->vfmt->alpha:%d"),video->vfmt->alpha);
	logging->out_debugf(_T("video->current_w:%d"),video->current_w);
	logging->out_debugf(_T("video->current_h:%d"),video->current_h);
#endif

#endif
}
